;; Generated package description from ssh-deploy.el  -*- no-byte-compile: t -*-
(define-package "ssh-deploy" "3.1.12.0.20201016.143950" "Deployment via Tramp, global or per directory." '((emacs "25")) :keywords '("tools" "convenience") :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :url "https://github.com/cjohansson/emacs-ssh-deploy")
